package apoc.warmup;

import apoc.util.TestUtil;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.neo4j.configuration.GraphDatabaseSettings;
import org.neo4j.test.rule.DbmsRule;
import org.neo4j.test.rule.ImpermanentDbmsRule;

import java.util.Arrays;
import java.util.List;
import java.util.Map;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotEquals;

/**
 * @author Sascha Peukert
 * @since 06.05.16
 */
public class WarmupTest {

    @Rule
    public DbmsRule db = new ImpermanentDbmsRule();

    @Before
    public void setUp() {
        TestUtil.registerProcedure(db, Warmup.class);
        // Create enough nodes and relationships to span 2 pages
        db.executeTransactionally("CREATE CONSTRAINT FOR (f:Foo) REQUIRE f.foo IS UNIQUE");
        db.executeTransactionally("UNWIND range(1, 300) AS i CREATE (n:Foo {foo:i})-[:KNOWS {bar:2}]->(m {foobar:3, array:range(1,100)})");
        // Delete all relationships and their nodes, but ones with the minimum and maximum relationship ids, so
        // they still span 2 pages
        db.executeTransactionally("MATCH ()-[r:KNOWS]->() " +
                "WITH [min(id(r)), max(id(r))] AS ids " +
                "MATCH (n)-[r:KNOWS]->(m) " +
                "WHERE NOT id(r) IN ids " +
                "DELETE n, m, r");
    }

    @Test
    public void testWarmup() {
        TestUtil.testCall(db, "CALL apoc.warmup.run()", r -> {
            assertEquals(4L, r.get("nodesTotal"));
            assertNotEquals(0L, r.get("nodePages"));
            assertEquals(2L, r.get("relsTotal"));
            assertNotEquals(0L, r.get("relPages"));
        });
    }

    @Test
    public void testWarmupProperties() {
        TestUtil.testCall(db, "CALL apoc.warmup.run(true)", r -> {
            assertEquals(true, r.get("propertiesLoaded"));
            assertNotEquals(0L, r.get("propPages"));
        });
    }

    @Test
    public void testWarmupDynamicProperties() {
        TestUtil.testCall(db, "CALL apoc.warmup.run(true,true)", r -> {
            assertEquals(true, r.get("propertiesLoaded"));
            assertEquals(true, r.get("dynamicPropertiesLoaded"));
            assertNotEquals(0L, r.get("arrayPropPages"));
        });
    }

    @Test
    public void testWarmupIndexes() {
        TestUtil.testCall(db, "CALL apoc.warmup.run(true,true,true)", r -> {
            assertEquals(true, r.get("indexesLoaded"));
            assertNotEquals( 0L, r.get("indexPages") );
        });
    }

    @Test
    public void testWarmupOnDifferentStorageEngines() {
        final List<String> supportedTypes = Arrays.asList("standard", "aligned");
        for (String storageType : supportedTypes) {
            db.restartDatabase(Map.of(GraphDatabaseSettings.db_format, storageType));
            TestUtil.registerProcedure(db, Warmup.class);

            TestUtil.testCall(db, "CALL apoc.warmup.run(true,true,true)", r -> {
                assertEquals(true, r.get("indexesLoaded"));
                assertNotEquals( 0L, r.get("indexPages") );
            });
        }
    }
}
